#include "DiametroFeret.h"
#include <cmath>
#include <QRgb>
#include <QImage>
#include <QFile>
#include <QTextStream>
#include <QString>
#include <QPoint>
#include <QList>

int colorObject = 255;
int colorBackground = 0;
float diametroMinFeret=0;
float diametroMaxFeret=0;
QString datos ="img;particula;minFeret;maxFeret;razon\n"; 

QPoint DiametroFeret::positionPixel(int neighbour)
{
	int x = 0;
	int y =0;

	if(neighbour==7)
	{
		x=2;  y=2;
	}
	else if(neighbour==6)
	{
		x=1; y=2;
	}
	else if(neighbour==5)
	{
		x=0;   y=2;
	}
	else if(neighbour==4)
	{
		x=0;   y=1;
	}
	else if(neighbour==3)
	{
		x=0;   y=0;
	}
	else if(neighbour==2)
	{
		x=1;   y=0;
	}
	else if(neighbour==1)
	{
		x=2;   y=0;
	}
	else if(neighbour==0)
	{
		x=2;  y=1;
	}

	return QPoint(x,y);
}

int DiametroFeret::positionNeighbour(int position_image, int position_mask, int height_or_width) 
{
	int delta=1;
	int position = abs(position_image - delta + position_mask);
  
	if(position>=height_or_width)
		position = 2*height_or_width - position -2;
  
	return position;
}

bool DiametroFeret::isObject(QImage &image, int x,int y) 
{
  return (qRed(image.pixel(x,y)) == colorObject);
}

bool DiametroFeret::isBackground(QImage &image, int x,int y) 
{
	return (qRed(image.pixel(x,y)) == colorBackground);
}

void DiametroFeret::pixelsRegion(QImage &image, QList<QPoint> *contourRegion, bool updateMaxMin)
{
	int xRegion=0;
	int yRegion=0;

	int xSearch =0;
	int ySearch =0;

	int xMin = image.width();
	int yMin = image.height();

	int xMax = 0;
	int yMax = 0;

	QPoint positionNeighbourPixel;
	for(int i=0;i<contourRegion->count();i++)
	{
		xRegion = (contourRegion->at(i)).x();
		yRegion = (contourRegion->at(i)).y();

		image.setPixel(xRegion,yRegion,colorBackground);

		//Se actualizan los valores minimo y el maximo
		if(updateMaxMin)
		{
			 if(xMin>xRegion)
				xMin = xRegion;

			 if(yMin>yRegion)
				yMin = yRegion;

			 if(xMax<xRegion)
				xMax = xRegion;

			 if(yMax<yRegion)
				yMax = yRegion;
		}

	 //Se hace la busqueda en cada uno de los 8 (0-7) vecios del pixel
	 // (xRegion,yRegion)
		for(int neighbour=0;neighbour<8;neighbour++)
		{
			positionNeighbourPixel = positionPixel (neighbour);

			xSearch =  positionNeighbour(xRegion,positionNeighbourPixel.x(),image.width());
			ySearch =  positionNeighbour(yRegion,positionNeighbourPixel.y(),image.height());

			if(isObject(image, xSearch,ySearch))
			{
				image.setPixel(xSearch,ySearch,colorBackground);
				contourRegion->append(QPoint(xSearch,ySearch));

			   //Se actualizan los valores minimo y el maximo
			   if(updateMaxMin)
			   {
					if(xMin>xRegion)
						xMin = xRegion;

					if(yMin>yRegion)
						yMin = yRegion;

					if(xMax<xRegion)
						xMax = xRegion;

					if(yMax<yRegion)
						yMax = yRegion;
			   }
			}
		}
	}

	if(updateMaxMin)
	{
			contourRegion->append(QPoint(xMin,yMin));
			contourRegion->append(QPoint(xMax,yMax));
	}
}

//Saca los puntos que corresponden a la region y calcula el minimo y el maximo punto :D
QList<QPoint> DiametroFeret::pixelsRegion(QImage &image,QPoint positionInitial,bool updateMaxMin)
{
	QList<QPoint> region;
	region.append(positionInitial);

	pixelsRegion(image,&region,updateMaxMin);
	return region;
}


int DiametroFeret::translation(int coordenada, int desplazamiento) 
{
	return coordenada+desplazamiento;
}

QPoint DiametroFeret::translation(QPoint coordenada, QPoint desplazamiento) 
{
	return coordenada+desplazamiento;
}

float DiametroFeret::distance( QPoint &point1,  QPoint &point2) 
{
	return static_cast<float>(sqrt(pow( static_cast<float>(point2.x()-point1.x()) , 2.0) + pow(static_cast<float>(point2.y()-point1.y()) , 2.0)));
}


   // este metodo actualiza los valores de las areas
QImage DiametroFeret::extractParticle(QList<QPoint> &particleList,QImage &image)
{
	QPoint minPixel = particleList.at(particleList.count()-2);        
	QPoint maxPixel = particleList.at(particleList.count()-1);

	particleList.removeAt (particleList.count()-1); //Se elimina el max
	particleList.removeAt (particleList.count()-1); //Se elimina el min

	int heightParticle = maxPixel.y()- minPixel.y()+1;
	int widthParticle = maxPixel.x()- minPixel.x()+1;

	QImage particle = QImage(widthParticle,heightParticle,QImage::Format_Indexed8);


	QVector<QRgb> lut(0,0);    
	for(int i=0;i<256;i++)
		lut.append(qRgb(i,i,i));
	particle.setColorTable(lut);       
	particle.fill(colorBackground);

	//Se extrae la region del objecto
    for(int k=0;k<particleList.count();k++)
    {
		int xPixel = (particleList.at(k)).x();
		int yPixel = (particleList.at(k)).y();

		particle.setPixel(translation(xPixel,-minPixel.x()),translation(yPixel,-minPixel.y()), colorObject);
		image.setPixel(xPixel,yPixel,colorBackground);
     }
	 
     return particle;
}



 float DiametroFeret::absValue(float number)
 {
     if(number<0)
         return  static_cast<float>(-1.0*number);
     
     return number;
 }





//CALCULO DE MIN Y MAX DE FERET

QPoint DiametroFeret::finalLinePerperdicular( QPoint &point1,  QPoint &point2,  float m) 
{
	float xPoint1 = static_cast<float>(point1.x());
	float yPoint1 = static_cast<float>(point1.y());
		
	float xPoint2 = static_cast<float>(point2.x());
	float yPoint2 = static_cast<float>(point2.y());
		
	float xPerpendicular = (((-m*xPoint2) + yPoint2 - (xPoint1 / m) - yPoint1) * m ) / -(1 +(m*m));
		
	float yPerpendicular = (m * xPerpendicular) - (m * xPoint2) + yPoint2;

		
	return QPoint(xPerpendicular,yPerpendicular);
	
}
	
	

QList<QPoint>  DiametroFeret::traceLineDiagonal(QImage &image, QPoint pointInitial, float tangent)
{
	int dx = static_cast<int>( static_cast<float>(image.width()-1) / 2.0 );
	int dy = static_cast<int>( static_cast<float>(image.height()-1) / 2.0 );
                     
	QPoint delta(dx,dy);
                     
	pointInitial = translation(pointInitial, -delta);
	                    
	int colorLine = 128;
	QList<QPoint> lineList;
                     
	if(absValue(tangent)>1.0)
	{                   
		tangent = 1.0/tangent;
		                   
		float x= static_cast<float>(pointInitial.x());
		int y = pointInitial.y();
			                  
		//mientras que se generen coordeandas validas, es decir no menotres a cero o mayores al taaño de la imagen
			//recordar q se trabaja con (alto/2 ,ancho/2) como centro de la imagen o punto (0,0)
		while(x<=dx && y<=dy && x >=-dx && y>=-dy)
		{                  
				//actual         
			int i = translation(std::floor(x),dx);
			int j =translation(y,dy);             
						               
			image.setPixel(i,j,colorLine);    
			      
			lineList<<QPoint(i,j);
			                  
				//siguiente      
			x = x + tangent;  
			y++;								      
		}
	}                   
	else                
	{                   
		int x=  pointInitial.x();
		float y = static_cast<float>(pointInitial.y());
                     
		while(x<=dx && y<=dy && x >=-dx && y >=-dy)
		{                  
			int i = translation(x,dx); 
			int j =translation(std::floor(y),dy);	
	              
                        image.setPixel(i,j,colorLine);
			lineList<<QPoint(i,j);
                     
			y = y + tangent;
			x++;
		}
	}
	
	return lineList;                
                     
}                    
                     
bool DiametroFeret::limitedParticle(QImage & image, QPoint pointInitial, float tangent)
{                    
	int dx = static_cast<int>( static_cast<float>(image.width()-1) / 2.0 );
	int dy = static_cast<int>( static_cast<float>(image.height()-1) / 2.0 );
                     
	QPoint delta(dx,dy);
                     
	pointInitial = translation(pointInitial, -delta);
	                    
	//int colorLine = 128;
	bool particle = false;
                     
	if(absValue(tangent)>1.0)
	{                   
		tangent = 1.0/tangent;
		                   
		float x= static_cast<float>(pointInitial.x());
		int y = pointInitial.y();
			                  
		//mientras que se generen coordeandas validas, es decir no menotres a cero o mayores al taaño de la imagen
			//recordar q se trabaja con (alto/2 ,ancho/2) como centro de la imagen o punto (0,0)
		while(x<=dx && y<=dy && x >=-dx && y>=-dy && !particle)
		{                  
				//actual         
			int i = translation(std::floor(x),dx);
			int j =translation(y,dy); 
			                  
			particle = isObject(image,i,j);
			                  
				//siguiente      
			x = x + tangent;  
			y++;								      
		}
	}                   
	else                
	{                   
		int x=  pointInitial.x();
		float y = static_cast<float>(pointInitial.y());
                     
		while(x<=dx && y<=dy && x >=-dx && y >=-dy && !particle)
		{                  
			int i = translation(x,dx); 
			int j =translation(std::floor(y),dy);	
			                  
			                  
			particle = isObject(image,i,j);			
			
			y = y + tangent;
			x++;
		}
	}
	
	return particle;
}


void DiametroFeret::calcularValoresFeret(QImage &image)
{
	QPoint pointInitial = QPoint(0,0);
	int height = image.height()-1;
	int width = image.width()-1;

	bool particle = false;
	int j=0;
	int i =0;
	QPoint limiteInf;
	QPoint limiteSup;
	QPoint finalPerperdicular;

	float distancia =0.0;

	QImage copy (image);


//Horizontal y Vertical ////////////////////////////////////
	float maxFeret = static_cast<float>(image.width());
	float minFeret =  static_cast<float>(image.height());


	if(image.height()>image.width())
	{
		maxFeret = static_cast<float>(image.height());
		minFeret = static_cast<float>(image.width());
	}


//Diagonal 1 y 5 /////////////////////////////////////////
	particle = false;
	j=height;

	while(!particle && j>0)
	{
		pointInitial = QPoint(0,j);
		particle=limitedParticle(image,pointInitial,2.0);
		j= j - 2;
	}
	traceLineDiagonal(copy,pointInitial,2.0);
	limiteSup = pointInitial;

	particle = false;
	i=width;

	while(!particle && i>0)
	{
		pointInitial = QPoint(i,0);
		particle=limitedParticle(image,pointInitial,2.0);
		i--;
	}
	traceLineDiagonal(copy,pointInitial,2.0);
	limiteInf = pointInitial;


	finalPerperdicular = finalLinePerperdicular(limiteSup,limiteInf,2.0);
	distancia = distance(limiteSup, finalPerperdicular);

	if(distancia> maxFeret)
		maxFeret = distancia;

	if(distancia< minFeret)
		minFeret = distancia;

// Diagonal 4 y 8 /////////////////////////////////////////////////
	particle = false;
	i=0;

	while(!particle && i<image.width())
	{
		pointInitial = QPoint(i,0);
		particle=limitedParticle(image,pointInitial,-2.0);
		i++;
	}
	traceLineDiagonal(copy,pointInitial,-2.0);
	limiteSup = pointInitial;


	particle = false;
	j=height;

	while(!particle && j>0)
	{
		pointInitial = QPoint(width,j);
		particle=limitedParticle(image,pointInitial,-2.0);
		j= j-2;
	}
	traceLineDiagonal(copy,pointInitial,-2.0);
	limiteInf = pointInitial;

	finalPerperdicular = finalLinePerperdicular(limiteSup,limiteInf,-2.0);
	distancia = distance(limiteSup, finalPerperdicular);

	if(distancia> maxFeret)
		maxFeret = distancia;

	if(distancia< minFeret)
		minFeret = distancia;


//Diagonal 3 y 7 ////////////////////////////////////////////////////////////////
	particle = false;
	j=0;

	while(!particle && j<image.height())
	{
		pointInitial = QPoint(0,j);
		particle=limitedParticle(image,pointInitial,-0.5);
		j++;
	}
	traceLineDiagonal(copy,pointInitial,-0.5);
	limiteSup = pointInitial;

	particle = false;
	i=width;

	while(!particle && i>0)
	{
		pointInitial = QPoint(i,height);
		particle=limitedParticle(image,pointInitial,-0.5);
		i--;
	}
	traceLineDiagonal(copy,pointInitial,-0.5);
	limiteInf = pointInitial;

	finalPerperdicular = finalLinePerperdicular(limiteSup,limiteInf,-0.5);
	distancia = distance(limiteSup, finalPerperdicular);

	if(distancia> maxFeret)
		maxFeret = distancia;

	if(distancia< minFeret)
		minFeret = distancia;


	//Diagonal 2y 6 ////////////////////////////////////////////////
	particle = false;
	j=height;

	while(!particle && j>0)
	{
		pointInitial = QPoint(0,j);
		particle=limitedParticle(image,pointInitial,0.5);
		j--;
	}
	traceLineDiagonal(copy,pointInitial,0.5);
	limiteInf= pointInitial;


	particle = false;
	i=width;

	while(!particle && i>0)
	{
		pointInitial = QPoint(i,0);
		particle=limitedParticle(image,pointInitial,0.5);
		i--;
	}

	traceLineDiagonal(copy,pointInitial,0.5);
	limiteSup= pointInitial;


	finalPerperdicular = finalLinePerperdicular(limiteSup,limiteInf,0.5);
	distancia = distance(limiteSup, finalPerperdicular);


	if(distancia> maxFeret)
		maxFeret = distancia;

	if(distancia< minFeret)
		minFeret = distancia;

//diagonal 1  45 //////////////////////////////////////////////////////////
	particle = false;
	j=0;

	while(!particle && j<image.height())
	{
		pointInitial = QPoint(0,j);
		particle=limitedParticle(image,pointInitial,-1.0);
		j++;
	}

	traceLineDiagonal(copy,pointInitial,-1.0);
	limiteSup = pointInitial;
//
	particle = false;
	i=width;

	while(!particle && i>0)
	{
		pointInitial = QPoint(i,height);
		particle=limitedParticle(image,pointInitial,-1.0);
		i--;
	}
	traceLineDiagonal(copy,pointInitial,-1.0);
	limiteInf = pointInitial;

	finalPerperdicular = finalLinePerperdicular(limiteSup,limiteInf,-1.0);
	distancia = distance(limiteSup, finalPerperdicular);


	if(distancia> maxFeret)
		maxFeret = distancia;

	if(distancia< minFeret)
		minFeret = distancia;

//diagonal 2 135 ///////////////////////////////////////
	particle = false;
	j=height;

	while(!particle && j>0)
	{
		pointInitial = QPoint(0,j);
		particle=limitedParticle(image,pointInitial,1.0);
		j--;
	}

	traceLineDiagonal(copy,pointInitial,1.0);
	limiteInf = pointInitial;
//
	particle = false;
	i=width;

	while(!particle && i>0)
	{
		pointInitial = QPoint(i,0);
		particle=limitedParticle(image,pointInitial,1.0);
		i--;
	}
	traceLineDiagonal(copy,pointInitial,1.0);
	limiteSup = pointInitial;

	finalPerperdicular = finalLinePerperdicular(limiteInf,limiteSup,1.0);
	distancia = distance(limiteInf, finalPerperdicular);


	if(distancia> maxFeret)
		maxFeret = distancia;

	if(distancia< minFeret)
		minFeret = distancia;

	diametroMinFeret= minFeret;
	diametroMaxFeret= maxFeret;

	//copy.save(dirOutput()+"/Diagonales.png");

	image = copy;
}


bool DiametroFeret::apply( QImage &imagenBinaria, QString nombreImagen)
{	
	int esfericas = 0; 
	int numParticula = 0;

	for(int y=0;y<imagenBinaria.height();y++)
	{
		for(int x=0;x<imagenBinaria.width();x++)
		{
			if(isObject(imagenBinaria, x,y))
			{
				//SE OBTIENEN LA PARTICULA DE LA IMAGEN
				QList<QPoint> objectList =  pixelsRegion(imagenBinaria,QPoint(x,y),true);
				
				
				QImage particle = extractParticle(objectList, imagenBinaria);								
				
				diametroMinFeret=0;
				diametroMaxFeret=0;
				calcularValoresFeret(particle);
				
				float razonFeret = diametroMinFeret/diametroMaxFeret;
				if(razonFeret > 0.5)
					esfericas++;
				datos+=nombreImagen+";Particula_"+QString::number(numParticula)+";"+QString::number(diametroMinFeret,'f',2)+";"+QString::number(diametroMaxFeret,'f',2)+";"+QString::number(razonFeret,'f',2)+"\n";
				
				qDebug()<<" Particula "<< numParticula<<" max Feret "<<diametroMaxFeret<<" min feret" <<diametroMinFeret<<" razon "<<razonFeret<<" tamahno "<<objectList.size();
				//particle.save(pathDirImagen+"\\imagen_"+nombreImagen+"-particula_"+QString::number(numParticula)+"_razonFeret_"+QString::number(razonFeret)+".png");
				numParticula++;
			}        	
		}
	}
	
	qDebug()<<" Esfericas "<<esfericas<<" total: "<<numParticula<<" mitad:"<<ceil(numParticula/2);
	if(esfericas > ceil(numParticula/2))
	{
		qDebug()<<" Esfericaaaaaaaaaaaaaaaa ";
		return true;
	}
	else
	{
		qDebug()<<" Obaladaaaaaaaaa";
		return false;
	}
}

void DiametroFeret::saveFile(const QString & fileName)
{
     QFile file(fileName);
     if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
     {
             qWarning("Error al escribir el logSalida.");
     }
     else
     {
       QTextStream out(&file);
       out<<datos;
       qDebug()<<"escribio bn logSalida";
     }

     file.close();
}

bool DiametroFeret::isCylindrical()
{
	QImage imagen("inFeret38");
	return apply( imagen, "prueba" );	
	//saveFile("/dos/Doctorado/TejidoEpitelial/feretFile/Resultados.csv");
}
