/***********************************************************************/
/*                                                                     */
/*   KMeans.cpp                                                        */
/*                                                                     */
/*                                                                     */
/*   Author: Claudia Mazo                                              */
/*   Date: 26.09.13                                                    */
/*                                                                     */
/*   Copyright (c) 2013  Claudia Mazo - All rights reserved            */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/

#include "KMeans.h"

KMeans::KMeans(const int numCentrIni)
	: numCentroidesIniciales(numCentrIni)
{
	
}

double KMeans::distancia(const Elemento& elem1, const Elemento& elem2)
{
	int r2= (elem1.r-elem2.r);
	int g2= (elem1.g-elem2.g);
	int s2= (elem1.s-elem2.s);
	
	return sqrt(r2*r2 + g2*g2 + s2*s2);
}

bool KMeans::iguales(const vector<Elemento> & vec1, const vector<Elemento> & vec2)
{
	if(vec1.size()!=vec2.size()) return false;
	typedef vector<Elemento>::const_iterator Iter;
	for(Iter it1=vec1.begin(), it2=vec2.begin(); it1!=vec1.end(); it1++, it2++)
		if( fabs(it1->r - it2->r) > 1e-9 ||	fabs(it1->g - it2->g) > 1e-9
		||	fabs(it1->s - it2->s) > 1e-9 )
			return false;
	
	return true;
}

void KMeans::procesar(CImg<int>& img1, CImg<int>& img2, vector< vector<char> >& etiquetas, vector<Elemento>& centroides)
{
	int numCentroides=numCentroidesIniciales;
	etiquetas = vector< vector<char> >(img1.width(), vector<char>(img1.height(), -1));

		centroides.push_back(Elemento(70,30,150)); //nucleos
		centroides.push_back(Elemento(160,70,30)); //resto
		centroides.push_back(Elemento(110,154,0)); //luz
	
	vector<Elemento> sumaElementos; //para el calculo de los nuevos centroides
	vector<int> numElementos;
	
	int numIteraciones=0;
	cerr<<"inicia";
	bool cambio=true;
	while(true)
	{
		cambio=false;
		sumaElementos=vector<Elemento>(numCentroides, Elemento());
		numElementos=vector<int>(numCentroides,0);
		
		for(int x=0;x<img1.width();x++)
		{
			for(int y=0;y<img1.height();y++)
			{
				Elemento elem(img1(x,y,0), img1(x,y,1), img2(x,y));
				double minDist = 255*255*4; 
				int imin=0;
				for(unsigned int i=0;i<centroides.size();i++)
				{
					double dist=distancia(elem,centroides[i]);
					if(dist<minDist)
					{
						minDist=dist;
						imin=i;
					}
				}
				
				sumaElementos[imin].r += elem.r;
				sumaElementos[imin].g += elem.g;
				sumaElementos[imin].s += elem.s;
				numElementos[imin]++;
				
				if(etiquetas[x][y]!=imin)
				{
					etiquetas[x][y]=imin;
					cambio=true;
				}
			}
		}
		cerr<<".";
		numIteraciones++;
		if(!cambio) break; //ningun punto fue cambiado de etiqueta (centroide)
		
		vector<Elemento>& nuevosCentroides=sumaElementos; 
		for(int i=numCentroides-1;i>=0; i--)
		{
			if(numElementos[i]==0)
			{
				nuevosCentroides.erase(nuevosCentroides.begin()+i); //un centroide que desaparece
				numCentroides--;
			}
			else
			{
				nuevosCentroides[i].r /= numElementos[i];
				nuevosCentroides[i].g /= numElementos[i];
				nuevosCentroides[i].s /= numElementos[i];
			}
		}
		//terminar = iguales(centroides,nuevosCentroides);
		centroides = nuevosCentroides;
	}
	cerr<<endl<<"Numero de Centroides: "<<centroides.size()<<endl;
	
	for(unsigned int i=0;i<centroides.size();i++)
	{
		cout<<"R="<<centroides[i].r<<" G="<<centroides[i].g<<" B="<<" S="<<centroides[i].s<<endl;
	
	}
	
	cerr<<"Numero de Iteraciones: "<<numIteraciones<<endl;

}


CImg<int> KMeans::segmentar(CImg<int> &img1, CImg<int> &img2, const char* n, const char* c, const char* l)
{
	vector< vector<char> > etiquetas;
	vector<Elemento> centroides;
	procesar(img1,img2, etiquetas, centroides);
	CImg<int> nucleos(img1.width(), img1.height()); //escala de grises
	CImg<int> luz(img1.width(), img1.height()); //escala de grises
	CImg<int> conectivo(img1.width(), img1.height()); //escala de grises
	CImg<int> nueva(img1.width(), img1.height(), 1, 3);
	
	nucleos.fill(0);
	luz.fill(0);
	conectivo.fill(0);
	
	for(int x=0; x<nueva.width(); x++)
		for(int y=0; y<nueva.height(); y++)
		{
			if(etiquetas[x][y]==0)
				nucleos(x,y)=255;
			else if(etiquetas[x][y]==1)
				conectivo(x,y)=255;
			else
				luz(x,y)=255;
				
			nueva(x,y,0)=color[etiquetas[x][y]].r;
			nueva(x,y,1)=color[etiquetas[x][y]].g;
			nueva(x,y,2)=color[etiquetas[x][y]].b;
			
			/*Elemento& centr=centroides[etiquetas[x][y]];
			nueva(x,y,0)= centr.r;
			nueva(x,y,1)= centr.g;
			nueva(x,y,2)= centr.s;*/
		}
	nucleos.save(n);
	conectivo.save(c);
	luz.save(l);
	return nueva;
}

//g++ -L/usr/lib/ -lpthread -lX11 KMeans.cpp -o pr -Wall -Dcimg_verbosity=3
//./pr example.png largest\ structure\ eigenvalues.jpg km.jpg
