/***********************************************************************/
/*                                                                     */
/*   RegionesGrandes.cpp                                               */
/*                                                                     */
/*                                                                     */
/*   Author: Claudia Mazo                                              */
/*   Date: 26.09.13                                                    */
/*                                                                     */
/*   Copyright (c) 2013  Claudia Mazo - All rights reserved            */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/

#include "RegionesGrandes.h"

int dx[] = { 0,-1, 0,+1};
int dy[] = {-1, 0,+1, 0};
int dxx[] = {-1,+1,-1,+1, 0,-1, 0,+1};
int dyy[] = {-1,-1,+1,+1,-1, 0,+1, 0};

RegionesGrandes::RegionesGrandes(CImg<int> img, int tar, int rep, int size, vector< list<Point> >& salida)
{
	original=img;
	image=original;
	target=tar;
	replace=rep;
	
	CImg<int> grandes(image.width(), image.height());
	grandes.fill(0);
	
	cimg_forXY(image, x, y)
		if( image(x,y) == target ){
			list<Point> region;
			
			floodfill(image, Point(x,y), region);
			int sz = region.size();
            if(sz>=size) {
				list<Point> bordes; 
				extraerBordes(original, region, bordes);
				salida.push_back(bordes);
				pintar(grandes, region, 128);
				pintar(grandes, bordes, 255);
			}
		}
    grandes.save("ff.pgm");

}

RegionesGrandes::RegionesGrandes(CImg<int> img, int tar, int rep, int size, CImg<int>& out)
{
	original=img;
	image=original;
	target=tar;
	replace=rep;
	
	CImg<int> grandes(image.width(), image.height());
	grandes.fill(0);
	
	cimg_forXY(image, x, y)
		if( image(x,y) == target ){
			list<Point> region;
			
			floodfill(image, Point(x,y), region);
			int sz = region.size();
			if(sz>=size)
				pintar(grandes, region, 128);
		}
	out=grandes;
	grandes.save("ff.pgm");
	//grandes.save(out);
}

bool RegionesGrandes::valid(const CImg<int>& img, int x, int y) {
	return x>=0 && x<img.width() && y>=0 && y<img.height();
}

void RegionesGrandes::floodfill(CImg<int>& img, Point inicio, list<Point>& salida) {
	queue<Point> q;

	q.push(inicio);
	salida.push_back(inicio);
	img(inicio.x,inicio.y)=replace;
	
	while(!q.empty()) {
		Point cur = q.front();
		q.pop();
		//if(img(cur.x,cur.y)==replace) continue;
		
		REP(dir, 8) {
			Point p(cur.x+dxx[dir],cur.y+dyy[dir]);
			if(valid(img, p.x, p.y) && img(p.x,p.y)==target) {
				img(p.x,p.y)=replace;
				salida.push_back(p);
				q.push(p);
			}
		}
	}
	//cout<<"salida termina en tamano "<<salida.size()<<endl;
}

void RegionesGrandes::pintar(CImg<int>& img, const list<Point>& puntos, int color) {
	FOREACH(p, puntos)
		img(p->x,p->y) = color;
}

void RegionesGrandes::extraerBordes(const CImg<int>& img, const list<Point>& puntos, list<Point>& salida) {
	FOREACH(p, puntos) {
		bool iguales=true;
		int clr = img(p->x,p->y);
		REP(dir,4) {
			int xx=p->x+dx[dir], yy=p->y+dy[dir];
			if(valid(img,xx,yy) && clr!=img(xx,yy)) {
				iguales=false;
				break;
			}
		}
		if(!iguales) {
			salida.push_back(*p);
		}
	}
}


